/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          pi version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var variable = '2';  // Variable selector: '1' => Angle of input line, '2' => Gap, '3' => 'Rotation', '4' => 'Line size' 

var iter = 1;   // Number of iterations for each variable value

// var i_angle;  Angle of input line 

// var g_angle;  The size of the gap between a fixed and a moving line

// var r_angle;  Rotation angle value

// var line_size;  Line size

var back_clr = '#fff';  // background canvas color during task

var pen_clr = '#000';  //  font color (base font) during the task 

var pen_clr_m = '#00f';  // font color for moving line 

var drag_clr = '#f00'; //  font color for a scrolling line when dragged with the mouse

var gap_type = '2';   // Gap type. range= [1..7]. see html select id="gap_type"

// var pen_width = 1;   // Pen (all lines) width

var is_circle = false;  // draw a outer circle around stimulus 
  
var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var prg_type = true;   //  Progress Bar is visible (true) or invisible (false)

var demo = false;     // if (demo == true) => Show task solutions 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 


// Other global constants and variables:

var item_cnt;       // number of items in the task. range [0..item_cnt-1]

var item;   // current item in task

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_storage= false; // local storage is enabled on user browser or not

  
//  global objects and arrays declarations..
 
var stim = {};   // stimulus object. see: stimGen()  

var middle = {};  // rectangular gap for the truncated part in the middle of the screen

var prop = {};  // stimulus properties objects. see setPar() => { 'angle': angle[i], 'gap': gap[i], 'rotation': rot[i], 'length': size[i] }; 

var dv = [];  // dependent variable for each item in range (error-pixels and percents). dimensions: [0 .. item_cnt-1] * [0 == error(px), 1 == percent]. 

var mis = [];  // missing values for each item in range: [0..item_cnt-1]
	
var slope = []; // fixed line slope (gradient) for each item [0..item_cnt-1] dy/dx= (y2 - y1) / (x2 - x1) => const

var err_max= [];  // maximum possible error in each task (px)

var r_max;    // radius within which all stimuli are drawn (scene). init on startTask() => setPar() 

var is_move= false, startX, startY, d_pos;  // dragging var. see: onMouseDown(), onMouseUp(), mouseMove(), pointValid()



/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */



// application init [check storage, set default values, show start panel..] 
function init() 
{ 
  setDefault();	      
  
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
	
	setIns();
	showStart();  
}

 
// set default values in 4 <textarea> objects => 4. variables
function setDefault()
{ 
	document.getElementById('i_angle').value= '45\n40\n50';
  document.getElementById('g_angle').value= '2\n3\n5';
	document.getElementById('r_angle').value= '0\n90\n180';
	document.getElementById('line_size').value= '100\n75\n50';
	/*   \r\n   or    \n  ?
	document.getElementById('i_angle').value= '45\r\n50\r\n55';
  document.getElementById('g_angle').value= '2\r\n3\r\n5';
	document.getElementById('r_angle').value= '0\r\n90\r\n180';
	document.getElementById('line_size').value= '100\r\n75\r\n50';  */
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (redundant?)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if ( document.getElementById('can').style.display == 'block' )  { 
    theEnd(true);	  // resize error			
	}
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('pi-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('pi-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		
	document.getElementById('ins_clr').style='color:' + pen_clr_m;
  document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';			
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'variable':
		  for (let i= 1; i<=4; i++)  {
				document.getElementById('fs_' + i).style.backgroundColor= ( e.value == i ) ? '#e5e5e5': '#fff'; 
			} 
		  break;
			
		case 'iter':
			if (e.validity.valid)  {
				msg.textContent= 'Total number of items = (number of iterations) * (number of entered valid values ​​for the selected variable)'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
				
    case 'back_clr': 		// background canvas color during task
		case 'pen_clr':  		//  font color (base font) during the task 
		case 'pen_clr_m': 	// font color for moving line 
		case 'drag_clr':		//  font color for a scrolling line when dragged with the mouse       
			msg.textContent= ( validColor(e.value) ) ? '' : 'The selected color is invalid.'; 
			break; 		
			
    case 'pen_width':
			if (e.validity.valid)  {
				msg.textContent= 'The larger the width of the pen, the larger and clearer all stimulus lines are. Recommended value is 1 (default).'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
			
		case 'is_circle':
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break;  	
		
    case 'prg_type':                                                                                                                                    
      msg.textContent= e.checked ? 'Yes. It will be displayed how many items are left to the end.' : 'No.';      
      break;    

		case 'demo': 
       msg.textContent= e.checked ? 'On. The solution of each task will be displayed (green line). The error size will be displayed.' : 'Off';  
       break; 
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity in most of options (except 4 textarea elements)
function optionsValid()  
{ 
  let res = document.getElementById('iter').validity.valid && document.getElementById('pen_width').validity.valid  &&  validColor(document.getElementById('back_clr').value) && validColor(document.getElementById('pen_clr').value) && validColor(document.getElementById('pen_clr_m').value) && validColor(document.getElementById('drag_clr').value);
  return res;              
}


// start task (research) 
function startTask(e)   
{      			
  initLayout(); 	
   
  setPar();
 
	stimGen();
	
	drawStim();    // draw first item (stimulus)
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  	
	document.getElementById('progress_holder').style.display = document.getElementById('prg_type').checked ? 'block' : 'none';
  document.getElementById('nav_items').textContent= '';	
	
  // canvas must be after set footer content => progress_holder visible or not
	let c= document.getElementById('can');
	c.style.display= 'block';   // see: pageResize() => ***
	c.width= window.innerWidth;  
	c.style.width= c.width + 'px';	     
	c.height= window.innerHeight - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight  
	c.style.height= c.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 			
}  
	
	
/* set internal application research parameters (global vars)
   prepare stimulus properties (input angle, gap, rotation angle, line size) for stimulus generator  */
function setPar()  
{  
	item = 0; 
	
	dv = [];
	mis = []; 	
	slope = [];
	err_max = [];
	
	stim = {};	
	middle = {};
	prop = {};
	
	validate();   // filter and validate variable values (4. textarea elements)
	
	// prepare stimulus properties (input angle, gap, rotation angle, line size) for stimulus generator
	
	let angle= [], gap= [], rot= [], size = [], v;
	
	// Variable selector: '1' => Angle of input line, '2' => Gap, '3' => 'Rotation', '4' => 'Line size' 
	
	// 1. angle of input line
	let lines = document.getElementById('i_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)  {
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			angle.push(v);
		}
	}	
	if  (variable == '1' && angle.length > 1) {
		angle= shuffle(angle);
	}
	
	// 2. Gap.
	lines = document.getElementById('g_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			gap.push(v);
		}
	}
	if  (variable == '2' && gap.length > 1) {
		gap= shuffle(gap);
	}
	
	// 3. Rotation angle
	lines = document.getElementById('r_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			rot.push(v);
		}
	}
	if  (variable ==  '3' && rot.length > 1) {
		rot= shuffle(rot);
	}
	
	// Line size
	lines = document.getElementById('line_size').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			size.push(v);
		}
	}
	if  (variable == '4' && size.length > 1) {
		size= shuffle(size);
	}
	
  switch (variable)
	{
		case '1': 
			item_cnt= angle.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'angle': angle[i], 'gap': gap[0], 'rotation': rot[0], 'length': size[0] }; 
			}
			break;
		case '2': 
			item_cnt= gap.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'angle': angle[0], 'gap': gap[i], 'rotation': rot[0], 'length': size[0] }; 
			}
			break;
		case '3': 
			item_cnt= rot.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'angle': angle[0], 'gap': gap[0], 'rotation': rot[i], 'length': size[0] }; 
			}			
			break;		
		case '4': 
			item_cnt= size.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'angle': angle[0], 'gap': gap[0], 'rotation': rot[0], 'length': size[i] }; 
			}			
			break;			
	}
  
	// after setting item_cnt
	for (let i= 0; i< item_cnt; i++) { 
    mis[i]= true;	
  }
}


// https://varun.ca/polar-coords/
// https://tutorial.math.lamar.edu/classes/calcii/polarcoordinates.aspx
// generate all stimuli for each item in range 0..item_cnt-1	
function stimGen()	
{			 	
	const off = 10;  // offset from edge
			// canvas position is set in end of initLayout() => after set footer content
	let c= document.getElementById('can'), w= c.width, h= c.height, x0= w/2, y0= h/2;				
	r_max = Math.round ( Math.min (w/2, h/2) - off);  // outer radius within this circle must be everything. circle center is (xs,ys) 
	
	for (let i = 0; i < item_cnt; i++)    // i => item  
	{			
			// 	prop[i] = { 'angle': angle[i], 'gap': gap[i], 'rotation': rot[i], 'length': size[i] }; 
		let i_angle= -degToRad(prop[i].angle); // for classic PI change sign of input angle
		let g_angle= degToRad(prop[i].gap);   
		let r_angle= degToRad(prop[i].rotation) + Math.PI/2;  // for classic PI set: + Math.PI/2;
		
    // 1. set borders of middle region (4. points)		
    let first =  {
			'begin' : {  
				x : x0 - r_max * Math.cos(r_angle-g_angle), // moving line - start point x
				y : y0 - r_max * Math.sin(r_angle-g_angle)  // moving line - start point y
			},
			'end' : {
				x : x0 - r_max * Math.cos(r_angle+g_angle), // fixed line - end point x
				y : y0 - r_max * Math.sin(r_angle+g_angle)  // fixed line - end point y					
			}
		};
		
		let	o_begin =  getOpposite(first.begin.x, first.begin.y, x0, y0);  // opposite x and y coord for outer point T(x_o, y_o)
		let   o_end =  getOpposite(first.end.x,   first.end.y,   x0, y0);  // opposite x and y coord for outer point T(x_o, y_o)
      
		let	second =  {
			'begin' : {  
				x : o_end.x,    // Note: reversed opposite begin <=> end
				y : o_end.y   
			},
			'end' : {
				x : o_begin.x,  // Note: reversed opposite begin <=> end
				y : o_begin.y   				
			}
		};
		
		let half = { 
      'first' : {
				x: divLine(first.begin.x, first.end.x, -1), 
        y: divLine(first.begin.y, first.end.y, -1)
			},
			'second' : {
				x: divLine(second.begin.x, second.end.x, -1), 
        y: divLine(second.begin.y, second.end.y, -1)
			}
		}
		
		middle[i] = {   
			'first'  : first,
      'second' : second,
			'half' : half 
		};    // see bellow: ***
    
    // 2. set fixed and moved (scroll) lines    	  
			
	  // 2a. set outer radius first
    let tmp1= getDistance (middle[i].first.begin.x,   middle[i].first.begin.y,  middle[i].first.end.x,  middle[i].first.end.y)/2;		
		let r_o = (r_max -  tmp1)  * prop[i].length / 100;       // remainder total radius - middle region width 
		prop[i].length_px = r_o;  // update percents into px
		prop[i].gap_px = tmp1;    // update angle arc into px
	  let offset = getDistance( middle[i].half.first.x,  middle[i].half.first.y, middle[i].first.begin.x, middle[i].first.begin.y );
		r_o += offset;
		let x_o = Math.round(x0 - r_o * Math.cos(i_angle + r_angle));  // fixed line - outer point x. 
		let y_o = Math.round(y0 - r_o * Math.sin(i_angle + r_angle));  // fixed line - outer point y. 
    let	opp_o = getOpposite(x_o, y_o, x0, y0); // opposite x and y coord for outer point T(x_o, y_o).  T(x0,y0) is center of canvas. 

    // 2b. compute inner points coord. intersection of outer line and 2. middle (gap) lines
    let p1 = lineIntersection( x_o, y_o,  opp_o.x, opp_o.y,  middle[i].first.begin.x, middle[i].first.begin.y,   middle[i].second.begin.x, middle[i].second.begin.y);   // outer radius line with 1. middle (gap) line intersection   
    let p2 = lineIntersection( x_o, y_o,  opp_o.x, opp_o.y,  middle[i].first.end.x,   middle[i].first.end.y,     middle[i].second.end.x,   middle[i].second.end.y);   // outer radius line with 2. middle (gap) line intersection   
		if ( getDistance(p1.x, p1.y, x_o, y_o) > getDistance(p2.x, p2.y, x_o, y_o) )  {  // which intersection (point) is closer to this outer radius? 
			let tmp = p1;  // swap points (coord)
			p1 = p2;
			p2 = tmp;
		}  //  point p1 is closer to T(x_o, y_o)

    let fixed = { 
			'inner' : { x : p1.x, y : p1.y },
			'outer' : { x : x_o, y : y_o } 
		};  // fixed.inner is closer to T(x_o,y_o)
		let length= getDistance(fixed.inner.x, fixed.inner.y, fixed.outer.x, fixed.outer.y);
				
		let correct = { 
			'inner' : { x : p2.x, y : p2.y },
			'outer' : { x : opp_o.x, y : opp_o.y } 
		};
			
		stim[i] = {
			'fixed': fixed,
			'correct' : correct 
		};		// moved line will be added latter	
    
		slope[i] = ( stim[i].fixed.outer.y - stim[i].fixed.inner.y ) / ( stim[i].fixed.outer.x - stim[i].fixed.inner.x );  // line slope or gradient: dy/dy
		
		// 3. add ofset for moved line (offset from correct slution)
		if (demo)  {  // equal as stim[i].correct (copy correct object)
			stim[i].moved = { 
				'inner' : { x : p2.x, y : p2.y },
				'outer' : { x : opp_o.x, y : opp_o.y } 
			};
		}
		else
		{		
			let theta = Math.atan(slope[i]), x, y, moved = {};  // moved => function result	
			let	d1 = getDistance ( middle[i].first.end.x + length * Math.cos(theta), middle[i].first.end.y + length * Math.sin(theta), correct.outer.x, correct.outer.y);
			let	d2 = getDistance ( middle[i].second.end.x + length * Math.cos(theta), middle[i].second.end.y + length * Math.sin(theta), correct.outer.x, correct.outer.y);
			if ( d1 > d2 )  
			{
				moved.inner = { x : middle[i].first.end.x, y : middle[i].first.end.y };
				let d = getDistance ( middle[i].first.end.x + length * Math.cos(theta+Math.PI), middle[i].first.end.y + length * Math.sin(theta+Math.PI), correct.outer.x, correct.outer.y);
				if ( d < d1 )  {
					theta+= Math.PI;
				}
				x = middle[i].first.end.x + length * Math.cos(theta);  // x = x0 + r * Math.cos(theta);
				y = middle[i].first.end.y + length * Math.sin(theta);  // y = y0 + r * Math.sin(theta);
			}	
			else  
			{
				moved.inner = { x : middle[i].second.end.x, y : middle[i].second.end.y };
				let d = getDistance ( middle[i].second.end.x + length * Math.cos(theta+Math.PI), middle[i].second.end.y + length * Math.sin(theta+Math.PI), correct.outer.x, correct.outer.y);							
				if ( d < d2 )  {  
					theta+= Math.PI;
				}
				x = middle[i].second.end.x + length * Math.cos(theta);  // x = x0 + r * Math.cos(theta);
				y = middle[i].second.end.y + length * Math.sin(theta);  // y = y0 + r * Math.sin(theta);
			}				
			moved.outer = { 'x' : x, 'y' : y };			
			stim[i].moved = moved;    // finally all done with scroll (moveable) line
		}  // if (!demo)	
		
		// *** add for Gap line type: part of input line => middle[i].gap_i		  
		if   ( getDistance( middle[i].first.begin.x,  middle[i].first.begin.y,   stim[i].fixed.inner.x, stim[i].fixed.inner.y ) 
			 <   getDistance( middle[i].second.begin.x, middle[i].second.begin.y,  stim[i].fixed.inner.x, stim[i].fixed.inner.y ) ) 
		{
			middle[i].gap_i= {
				smaller :  { 'x': middle[i].first.begin.x,  'y': middle[i].first.begin.y },      // smaller distance to stim[i].fixed.inner.x
			  greater :  { 'x': middle[i].second.begin.x,  'y': middle[i].second.begin.y } };  // greater distance to stim[i].fixed.inner.x
		}
		else  {		
			middle[i].gap_i= { 
			  smaller :  { 'x': middle[i].second.begin.x,  'y': middle[i].second.begin.y },
				greater :  { 'x': middle[i].first.begin.x,  'y': middle[i].first.begin.y } }			 
		}		
		
		// calculate maximum error
		let d1= getDistanceFromLine( stim[i].correct.inner.x, stim[i].correct.inner.y, stim[i].correct.outer.x, stim[i].correct.outer.y, middle[i].first.end.x, middle[i].first.end.y );
		let d2= getDistanceFromLine( stim[i].correct.inner.x, stim[i].correct.inner.y, stim[i].correct.outer.x, stim[i].correct.outer.y, middle[i].second.end.x, middle[i].second.end.y );
			
		err_max[i]= Math.max(d1,d2);
    
		if (demo)  {		
			dv[i]= [0, 0];
    }
    else  {
			dv[i]= [null,null];	  // dv[i]= [err_max[i], 100]; 
		}
		
	}  // for (let i = 0; i < item_cnt; i++)

}


// convert degrees (deg) to radians
function degToRad(deg)  {
	return (deg * Math.PI/180);   
}


/*  convert radians (rad) to degrees
function radToDeg(rad)  {
	return (rad * 180 / Math.PI); 
} */
 
 
// isNaN("") = false: the empty string is converted to 0 which is not NaN                   
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 


function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}


// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


// https://www.w3.org/2002/09/tests/keys.html	 
// https://stackoverflow.com/questions/17164239/add-delete-and-arrow-key-into-regular-expression 
// https://stackoverflow.com/questions/13196945/keycode-values-for-numeric-keypad/13196983
// 4. variables => basic textarea validator => onkeydown() => see html tag
function updateVar(event)
{
	event = event || window.event;
	let key = event.keyCode || event.which;
	if( key == 13 || key == 37 || key == 38 || key == 39 || key == 40 || key == 8 || key == 46 ) { // ENTER, Left / Up / Right / Down Arrow, Backspace, Delete keys
		return;  // Don't validate the input if ENTER, arrows, delete and backspace keys were pressed 
	}
	let is_digit = (key >= 48 && key <= 57) || (key >= 96 && key <= 105);
	if( !is_digit )  {
	event.returnValue = false;
	if(event.preventDefault) 
		event.preventDefault();
	}   
}


// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	


/* when user move blue circle (dot)
   canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
   this event is fired by html canvas element => document.getElementById('can') 
	 <canvas id='can' onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>  */
function mouseMove(e)
{   
	if (is_move)   
	{ 	
		let xm= e.offsetX, ym= e.offsetY;
    if ( pointValid(xm, ym)  ) {  
		
			// update scrolling line position => https://stackoverflow.com/questions/5559248/how-to-create-a-draggable-line-in-html5-canvas
			let dx= xm - startX;  // calc how far mouse has moved since last mousemove event
			let dy= ym - startY;
    
			// update scrolling line position
			stim[item].moved.inner.x += dx;  	// change line vertices by distance moved
			stim[item].moved.inner.y += dy;
			stim[item].moved.outer.x += dx;
			stim[item].moved.outer.y += dy;	
			startX= xm;
			startY= ym;

			// update dependent variable for this item => mouse distance from correct solution (line)
			dv[item][0]= getDistanceFromLine( stim[item].correct.inner.x, stim[item].correct.inner.y, stim[item].correct.outer.x, stim[item].correct.outer.y, xm, ym );	
			
			mis[item]= false;	// this item is not missing
			
			drawStim();  // repaint stimulus	
		}
	}		 
}


// x,y => mouse positions. Note: x = e.offsetX == e.clientX; y = e.offsetY == e.clientY 
// test new coord: T(stim[item].moved.inner.x + dx, stim[item].moved.inner.y + dy) 
function pointValid(x,y) 
{
	    
		let c= document.getElementById('can'), x0= Math.round(c.width/2), y0= Math.round(c.height/2);
	// https://stackoverflow.com/questions/5559248/how-to-create-a-draggable-line-in-html5-canvas
    let dx= x - startX;  // calc how far mouse has moved since last mousemove event
    let dy= y - startY;

		// test new coord: T(stim[item].moved.inner.x + dx, stim[item].moved.inner.y + dy)    		
		if ( getDistance(stim[item].moved.inner.x + dx, stim[item].moved.inner.y + dy, x0, y0) > r_max ) {
			 //stopDrag(); 
			 return false; 
		}
		
    let d = getDistanceFromLine( middle[item].half.first.x, middle[item].half.first.y,  
		                             middle[item].half.second.x, middle[item].half.second.y,
																 stim[item].moved.inner.x + dx, stim[item].moved.inner.y + dy ); // getDistanceFromLine(x1, y1, x2, y2, x0, y0)
    let d_min = getDistance( middle[item].half.first.x,  middle[item].half.first.y, middle[item].first.begin.x, middle[item].first.begin.y );     
		if (d < d_min-1)  {		 	
			return false; 
		}
		
		let p1= lineIntersection( middle[item].first.begin.x, middle[item].first.begin.y,  middle[item].second.begin.x,  middle[item].second.begin.y, stim[item].moved.inner.x+dx, stim[item].moved.inner.y+dy, stim[item].moved.outer.x+dx, stim[item].moved.outer.y+dy );
		let p2= lineIntersection( middle[item].first.end.x, middle[item].first.end.y, middle[item].second.end.x, middle[item].second.end.y, stim[item].moved.inner.x+dx, stim[item].moved.inner.y+dy, stim[item].moved.outer.x+dx, stim[item].moved.outer.y+dy ); 
		
		let d_begin = getDistance( p1.x, p1.y,  stim[item].moved.inner.x+dx, stim[item].moved.inner.y+dy);  
		let d_end =   getDistance( p2.x, p2.y,  stim[item].moved.inner.x+dx, stim[item].moved.inner.y+dy); 
		if ( Math.sign (d_begin - d_end) != d_pos) {
			return false;
		}
		
	  return true;
}


/* https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
  If the line passes through two points P1 = (x1, y1) and P2 = (x2, y2) then the distance of (x0, y0) from the line is d
  d = distance = (P1, P2, (x0,y0)) 
	d = | (x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1) | / ( SQRT ( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) )  */
function getDistanceFromLine(x1, y1, x2, y2, x0, y0)
{
	let d = Math.abs( (x2-x1) * (y1-y0) - (x1-x0) * (y2 - y1) ) / Math.sqrt( Math.pow(x2-x1, 2) + Math.pow(y2-y1, 2) );  
	return d;
}	


// html <canvas id='can' onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                         
function mouseDown(e)  
{	 
	let xm= e.offsetX, ym= e.offsetY;   // mouse pointers. Note: e.offsetX == e.clientX;  e.offsetY == e.clientY 
	let d= getDistanceFromLine( stim[item].moved.inner.x, stim[item].moved.inner.y, stim[item].moved.outer.x, stim[item].moved.outer.y, xm, ym );   // get mouse distance from moved line. e.offsetX and e.offsetY => mouse coordinates x, y	
	let ok= d <= 10 ? true: false;  // 1. condition
	if ( ok )  {  // and 2. condition
		let xc= divLine(stim[item].moved.inner.x, stim[item].moved.outer.x, -1);			
		let yc= divLine(stim[item].moved.inner.y, stim[item].moved.outer.y, -1);
		let d = getDistance(xm, ym, xc, yc);  // mouse distance from center of moving line 	
		let length= getDistance(stim[item].moved.inner.x, stim[item].moved.inner.y, stim[item].moved.outer.x, stim[item].moved.outer.y);		
		ok = d <= length/2 ? true : false;		
	}
	if ( ok )  {				
		let p1= lineIntersection( middle[item].first.begin.x, middle[item].first.begin.y,  middle[item].second.begin.x,  middle[item].second.begin.y, stim[item].moved.inner.x, stim[item].moved.inner.y, stim[item].moved.outer.x, stim[item].moved.outer.y );
		let p2= lineIntersection( middle[item].first.end.x, middle[item].first.end.y, middle[item].second.end.x, middle[item].second.end.y, stim[item].moved.inner.x, stim[item].moved.inner.y, stim[item].moved.outer.x, stim[item].moved.outer.y ); 
		let d_begin = getDistance( p1.x, p1.y,  stim[item].moved.inner.x, stim[item].moved.inner.y);  
		let d_end =   getDistance( p2.x, p2.y,  stim[item].moved.inner.x, stim[item].moved.inner.y); 
		d_pos= Math.sign (d_begin - d_end);     // only sign is important
		startX= xm;
		startY= ym;
		
		is_move= true;
		document.getElementById('can').style.cursor= 'move';  //  'ew-resize'  or  'ns-resize'  
    
		drawStim();		
  }
	else  {
		stopDrag();
	}
}


// html <canvas id='can' onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                        
function mouseUp()  
{
	stopDrag();
  drawStim();
}


// stop dragging moving line
function stopDrag()  
{
	is_move= false;
  document.getElementById('can').style.cursor= 'auto';     //c.style.cursor= 'initial';  // because of IE11
	drawStim();	
}


// draw (repaint) current stimulus           
function drawStim() 
{    		
	
	let c= document.getElementById('can'), ctx= c.getContext('2d'), w= c.width, h= c.height;		
  ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, w, h);    // ctx.clearRect(0, 0, w, h); 
	
	ctx.lineWidth= pen_width;   	
	
  ctx.beginPath(); 
	ctx.strokeStyle= is_move ? drag_clr : pen_clr_m; 
	ctx.moveTo(stim[item].moved.inner.x, stim[item].moved.inner.y);  // move to inner point of moved (scroll) line     
	ctx.lineTo(stim[item].moved.outer.x, stim[item].moved.outer.y);  // draw line to outer point of moved (scroll) line 
	ctx.stroke();  

	ctx.strokeStyle= pen_clr; 
	
	ctx.beginPath();
	ctx.moveTo(stim[item].fixed.inner.x, stim[item].fixed.inner.y);  // move to inner point of fixed line     
	ctx.lineTo(stim[item].fixed.outer.x, stim[item].fixed.outer.y);  // draw line to outer point of fixed line 
  ctx.stroke(); 

	switch ( gap_type )			// draw gap
	{
		case '1':  // space gap between the 2 lines in the middle will have the color of a base pencil
			ctx.beginPath();	
			ctx.moveTo(middle[item].first.begin.x,  middle[item].first.begin.y);
			ctx.lineTo(middle[item].first.end.x,    middle[item].first.end.y);
			ctx.lineTo(middle[item].second.end.x,   middle[item].second.end.y);
			ctx.lineTo(middle[item].second.begin.x, middle[item].second.begin.y);	
			//ctx.lineTo(middle[item].first.begin.x, middle[item].first.begin.y);
			ctx.closePath();
			ctx.fillStyle= pen_clr;  // only tmp
			ctx.fill();
			ctx.fillStyle= back_clr; // restore canvas background-color
			break;
		
		case '2': // both gap lines are visible. default			
			ctx.beginPath();
			ctx.moveTo(middle[item].first.begin.x,   middle[item].first.begin.y);
			ctx.lineTo(middle[item].second.begin.x,  middle[item].second.begin.y);
			ctx.moveTo(middle[item].first.end.x,     middle[item].first.end.y);
			ctx.lineTo(middle[item].second.end.x,    middle[item].second.end.y);
			ctx.stroke(); 
			break;
			
		case '3':  // smaller part of gap line along the fixed line and gap line along the moving line is visible
		case '4':	 // bigger part of gap line along the fixed line and gap line along the moving line is visible		
			ctx.beginPath();
			ctx.moveTo(stim[item].fixed.inner.x, stim[item].fixed.inner.y);
			if (gap_type == '3')  {
				ctx.lineTo(middle[item].gap_i.smaller.x, middle[item].gap_i.smaller.y);
			}
			else {
				ctx.lineTo(middle[item].gap_i.greater.x, middle[item].gap_i.greater.y);	
			}
			ctx.moveTo(middle[item].first.end.x, middle[item].first.end.y);
			ctx.lineTo(middle[item].second.end.x, middle[item].second.end.y)
			ctx.stroke();
			break; 	
		          
		case '5': // only the gap line along the fixed line is visible			
			ctx.beginPath();
			ctx.moveTo(middle[item].first.begin.x,  middle[item].first.begin.y);
			ctx.lineTo(middle[item].second.begin.x, middle[item].second.begin.y);
			ctx.stroke(); 
			break;	
	
		case '6': // only the gap line along the moving line is visible
			ctx.beginPath();
			ctx.moveTo(middle[item].first.end.x,  middle[item].first.end.y);
			ctx.lineTo(middle[item].second.end.x, middle[item].second.end.y);
			ctx.stroke();
			break; 
			
		//default:  // or case '7': => both gap lines are invisible
			 // do nothing;  			
	}
	
	if (is_circle)  {
		ctx.beginPath();
		ctx.arc(w/2, h/2, r_max, 0, 2*Math.PI);
		ctx.stroke();
	}
 
	if (demo)  
	{	
		// draw correct (fixed extension) line inside the gap in the middle 
		ctx.beginPath(); 
		ctx.moveTo(stim[item].fixed.inner.x,   stim[item].fixed.inner.y);   		// move to inner point of fixed line    
		ctx.lineTo(stim[item].correct.inner.x, stim[item].correct.inner.y);    //  draw line to inner point of correct line 
		ctx.stroke();
				
		// draw the correct line (solution)
		ctx.strokeStyle= 'darkgreen';   // fixed color
		ctx.beginPath(); 
		ctx.moveTo(stim[item].correct.inner.x, stim[item].correct.inner.y);  // move to inner point of moved (scroll) line     
		ctx.lineTo(stim[item].correct.outer.x, stim[item].correct.outer.y);  // draw line to outer point of moved (scroll) line 
		ctx.stroke();		
		  //ctx.strokeStyle= pen_clr;			
	
		let range = err_max[item];  	
		let percent =  100 * ( ( range -  Math.abs ( range - dv[item][0] ) ) / range ); 
    dv[item][1]= percent; 			// update dependent variable (percents)
		
		// https://jsfiddle.net/dwahlin/GDTBy/		
		ctx.fillStyle= pen_clr;  //ctx.fillStyle= '#000';
		ctx.font = '1em Arial';   
		ctx.font.fontFamily = 'Arial, Verdana, Helvetica, sans-serif';  // https://www.w3schools.com/cssref/css_websafe_fonts.asp
		ctx.textAlign = 'middle';
		ctx.textBaseline= 'bottom';				
		ctx.save();
		let xs= stim[item].moved.inner.x; // divLine(stim[item].moved.inner.x, stim[item].moved.outer.x, -1);			
		let ys= stim[item].moved.inner.y; // divLine(stim[item].moved.inner.y, stim[item].moved.outer.y, -1);
		ctx.translate(xs,ys);
		ctx.rotate(Math.atan(slope[item]));	
		ctx.fillText('Error= ' + dv[item][0].toFixed(2) + ' (' + percent.toFixed(2) + '%)', 0, 0);
		//ctx.translate(-xs,-ys);
		ctx.restore(); 			
	}		
	
	updateFooter();	
}


/* returns the opposite point around the circle of rotation 
   x, y = requested point coordinates
   T(x0,y0) is center of rotation  */
function getOpposite(x,y,x0,y0)
{
	let angle = Math.atan2( y-y0, x-x0 ) + Math.PI;		
	let d= getDistance(x, y, x0, y0); 
	let res = { x : Math.round (x0 + d * Math.cos(angle) ), y : Math.round (y0 + d * Math.sin(angle) ) };
	return res;
}	


/* https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection#Given_two_points_on_each_line
   https://www.geeksforgeeks.org/program-for-point-of-intersection-of-two-lines/
   https://www.cuemath.com/geometry/intersection-of-two-lines/
	 return intersection of 2 lines given by 2 points
	 (x1,y1) and (x2,y2) => first line coord
	 (x3,y3) and (x4,y4) => second line coord  */
function lineIntersection(x1,y1, x2,y2, x3,y3, x4, y4)
{
	let d = ( x1 - x2 ) * (y3 - y4) - (y1 - y2) * (x3 - x4);
	if ( d == 0 )  { 
		return null; 
	}
	else  {
		let x = ( (x1 * y2 - y1 * x2 ) * (x3 - x4) - (x1 - x2) * (x3 * y4 - y3 * x4) ) / d;
		let y = ( (x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (x3*y4 - y3 * x4) ) / d;
		let res = { 'x' : Math.round(x), 'y': Math.round(y) };  //let res = { 'x' : x, 'y': y };
		return res;
	}	
}

  
// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	variable = document.getElementById('variable').value;

	if (document.getElementById('iter').validity.valid)  {  
	  iter = Math.round(parseInt(document.getElementById('iter').value, 10));			
  } 	
	
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr').value))  {		
		pen_clr= document.getElementById('pen_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr_m').value))  {		
		pen_clr_m= document.getElementById('pen_clr_m').value;
  }
	
  if (validColor(document.getElementById('drag_clr').value))  {		
		drag_clr= document.getElementById('drag_clr').value;
  }		
	
	if (document.getElementById('gap_type').selectedIndex > -1 )  {
		gap_type= document.getElementById('gap_type').value;
	}
	
	if (document.getElementById('pen_width').validity.valid)  {  
	  pen_width = Math.round(parseInt(document.getElementById('pen_width').value, 10));			
  }

	is_circle= document.getElementById('is_circle').checked ? true : false;
  
	is_back= document.getElementById('is_back').checked ? true : false;      
  	
	prg_type= document.getElementById('prg_type').checked ? true : false;   
		
  demo= document.getElementById('demo').checked ? true : false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// Validates and filters values entered in 4 lists (textarea) of variables
// function call by user (validate button) or setPar() at startTAsk() 
function validate()
{
	// Variable selector: '1' => Angle of input line, '2' => Gap, '3' => 'Rotation', '4' => 'Line size'
	
	let angle= [], gap= [], rot= [], size = [], v;
	
	// 1. angle of input line. Enter only one integer value in the range of 30-60 per line. Default value = 45.
	let lines = document.getElementById('i_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);  // code here using lines[i] which will give you each line
		if ( isInt(lines[i]) )  { 
		  v = parseInt(lines[i], 10);
			if ( v >= 30  &&  v <= 60 )  {
				angle.push(v);
			}
		}
	}
	if (angle.length == 0)  {
		angle.push(45);
	}	 	
	document.getElementById('i_angle').value= (variable == '1') ? angle.join('\n') : angle[0];
	//document.getElementById('i_angle').value= angle.join('\n');
	
	// 2. Gap. Enter only one integer value in the range of 1-25 per line. Default value = 3
	lines = document.getElementById('g_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  
		  v = parseInt(lines[i], 10);
			if ( v >= 1  &&  v <= 25 )  {
				gap.push(v);
			}
		}
	}
	if (gap.length == 0)  {
		gap.push(3);
	}	
	document.getElementById('g_angle').value= (variable == '2') ? gap.join('\n') : gap[0];
  // document.getElementById('g_angle').value= gap.join('\n');
	
	// 3. Rotation angle. Enter only one integer value in the range of 0-180 per line. Default value = 0. 
	lines = document.getElementById('r_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  //code here using lines[i] which will give you each line
		  v = parseInt(lines[i], 10);
			if ( v >= 0  &&  v <= 180 )  {
				rot.push(v);
			}
		}
	}
	if (rot.length == 0)  {
		rot.push(0);
	}	
	document.getElementById('r_angle').value= (variable == '3') ? rot.join('\n') : rot[0];
	// document.getElementById('r_angle').value= rot.join('\n');
	
	// 4. Line size. Enter only one integer value in the range of 50-100 per line. Default value = 100
	lines = document.getElementById('line_size').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  //code here using lines[i] which will give you each line
		  v = parseInt(lines[i], 10);
			if ( v >= 50  &&  v <= 100 )  {
				size.push(v);
			}
		}
	}
	if (size.length == 0)  {
		size.push(100);
	}	
	document.getElementById('line_size').value= (variable == '4') ? size.join('\n') : size[0];
	// document.getElementById('line_size').value= size.join('\n');
	
	if (!document.getElementById('iter').validity.valid)  {  
	  document.getElementById('iter').value= iter;		// restore last valid iteration value	
  } 	
	
	if (!document.getElementById('pen_width').validity.valid)  {  
	  document.getElementById('pen_width').value= pen_width;		// restore last valid pen width value	
  } 	
	
}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd(false);  // end of task if item >= item_cnt 
  }
}


/* divide line 
  first => first point (P1) 
  second => second point (P2)
  n => (distance(P1-Q)/ distance(Q-P2)) = lambda = -m/n = negative ratio length: P1-Q/Q-P2
  use n= -1 for half divide line (50%) 
	returns: coordinate of point (Q) between first and second point   */   
function divLine(first,second,n)  {     
  return (first-n*second)/(1-n);   // Math.round((first-n*second)/(1-n));
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                        
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
	}
}


// end of task (research)
// if (is_error== true) => page (window) was resized during task. Show only html error warning
function theEnd(is_error)  
{     
	
	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('can').style.display= 'none';   // must be. see: pageResize() call theEnd() if ( document.getElementById('can').style.display == 'block' ) theEnd(true)
  //document.getElementById('footer').style.display= 'none'; 
 
	if (is_error) {  // Show only html error warning
		document.getElementById('error').style.display= 'block';
		document.getElementById('result_wrapper').style.display= 'none';
		document.getElementById('score').textContent= '';
		return;
	}
	
  document.getElementById('error').style.display= 'none';
	document.getElementById('result_wrapper').style.display= 'block';
	
	const D = 3;  // number of decimals
  let blank= 0, rows = [], header = [], var_name, tmp; 
	let err_v = [], per_v = [], var_v = [], var_g= {};  // non-missing error (px), non-missing error (%), non-missing variable values (interval scale), var_g => group by variable value in px   			

  switch (variable) {
		case '1': 
		  var_name= 'Angle';
			break;
		case '2': 
		  var_name= 'Gap';
			break;	
		case '3': 
		  var_name= 'Rotation';
			break;
    case '4': 
		  var_name= 'Line size';
			break;  			
  }	
	
	if (variable == '1' || variable == '3')  { 
	  header= [ 'Item', var_name,                   'Error-px', 'Error %', 'Moved' ];
	}
	else {
		header= [ 'Item', var_name, var_name + '-px', 'Error-px', 'Error %', 'Moved' ];
	}
	for (let i = 0; i < item_cnt; i++)  
	{	
		// Variable selector: '1' => Angle of input line, '2' => Gap, '3' => 'Rotation', '4' => 'Line size'   		
		switch (variable)  {				
			case '1':				
				var_v.push (prop[i].angle);
				tmp= prop[i].angle;   
				break;
			case '2':				
				var_v.push (prop[i].gap_px);  // gap in pixels. set in stimGen()
				tmp = prop[i].gap;				// string 2. column in table1 (variable value selected by user)
				break;
			case '3':
				var_v.push ( prop[i].rotation );
				tmp= prop[i].rotation;
				break;
			case '4':
				var_v.push ( prop[i].length_px ); // line length in pixels. set in stimGen()
				tmp= prop[i].length;         // string 2. column in table1 (variable value selected by user)
				break;					
		}		
		let last = var_v[var_v.length-1];
		// for decriptive statistics set group by variable value (table2 + table3)..
		let key = last.toFixed(D);   // key => last one pushed into var_v
		if ( typeof var_g[key]  === 'undefined' )  {
			var_g[key] = {   
				'error'  : [],
				'percent' : []
			}; 
		}

    if ( mis[i] == true )  
		{
			blank++;	
		  if (variable == '1' || variable == '3')  { 		//  angle + rotation 	== 5. elements 				   			
				rows[i]= [ parseInt(i+1,10), tmp,  '', '', 'No' ];  //=>['Item', 'Variable', 'Error-px',   'Error %',          'Moved'] 
			}
			else  {  // gap + line_size	=> 6. elements	   			
				rows[i]= [ parseInt(i+1,10), tmp,  last.toFixed(D),  '', '', 'No' ];   // => [ 'Item', 'Variable', 'Variable-px'          'Error-px',             'Error %', 'Moved' ]  
			}
		}
		else  
		{
			if (!demo) {  // compute percents. if (demo) => dv[i][1] is computed in drawStim() 
				let range = err_max[i];  	
				dv[i][1]= 100 * ( ( range -  Math.abs ( range - dv[i][0] ) ) / range );
			}
		  err_v.push ( dv[i][0] );  // error in pixels
		  per_v.push ( dv[i][1] );  // error in percents

			if (variable == '1' || variable == '3')  { //=>['Item', 'Variable', 'Error-px',   'Error %',          'Moved']  			 				   			
				rows[i]= [ parseInt(i+1,10), tmp,                         dv[i][0].toFixed(D), dv[i][1].toFixed(D), 'Yes' ]; 
			}
			else  {  // => [ 'Item', 'Variable', 'Variable-px'          'Error-px',             'Error %', 'Moved' ]   				   			
				rows[i]= [ parseInt(i+1,10), tmp,  last.toFixed(D),   dv[i][0].toFixed(D), dv[i][1].toFixed(D), 'Yes' ]; 
			}
			
			var_g[key].error.push( dv[i][0] );    // error in pixels => err_v
			var_g[key].percent.push( dv[i][1] );  // error in percents => per_v
		}
	}  // for (let i = 0; i < item_cnt; i++)
		
  createTableArray(document.getElementById('table1'), rows, header); 


	// calculate Descriptive statistics grouped by variable (px) and total => table2
	//document.getElementById('table2_title').textContent= (variable == 1 || variable == 3) ? 'Descriptive statistics grouped by variable ' + var_name + ' and total:' : 'Descriptive statistics grouped by variable ' + var_name  + '-px and total:'; 
	document.getElementById('table2_title').textContent= 'Descriptive statistics (px) grouped by variable ' + var_name + ' and Total:'; 
	let arr, n, m, sd;
	rows = [];  // reset table rows (for table2)
	for (let key in var_g)  {
		arr= var_g[key].error;  // grouped by variable for this key
		n= arr.length;
		m =  ( n > 0 ) ? Stat.m(arr).toFixed(D) : '';     //  Errors (px). mean only for non-missing items in task
		sd = ( n > 1 ) ? Stat.sd(arr, m).toFixed(D): '';  // Errors (px). standard deviation only for non-missing items in task  	 	
		rows.push( [key, m, sd, n] );  
	}
	// totals (px)
	n= err_v.length;
	m =  ( n > 0 ) ? Stat.m(err_v).toFixed(D) : '';     //  Total Errors (px). mean only for non-missing items in task
	sd = ( n > 1 ) ? Stat.sd(err_v, m).toFixed(D): '';  // Total Errors (px). standard deviation only for non-missing items in task  	 	
  rows.push( ['Total (px)', m, sd, n] );  // last row
	let data = [ m, sd ];  // index 0 => mean for px, index 1 => stand. deviation for px
	header= [var_name, 'Average', 'Standard Deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 
	
	
	// calculate Descriptive statistics grouped by variable (%) and total => table3
	document.getElementById('table3_title').textContent= 'Descriptive statistics (%) grouped by variable ' + var_name + ' and Total:';
	rows = [];  // reset table rows (for table3)
	for (let key in var_g)  {
		arr= var_g[key].percent;  // grouped by variable for this key
		n= arr.length;
		m =  ( n > 0 ) ? Stat.m(arr).toFixed(D) : '';     //  Errors (%). mean only for non-missing items in task
		sd = ( n > 1 ) ? Stat.sd(arr, m).toFixed(D): '';  //  Errors (%). standard deviation only for non-missing items in task  	 	
		rows.push( [key, m, sd, n] );  
	}
  // totals	(%)
  n= per_v.length;
	m =  ( n > 0 ) ? Stat.m(per_v).toFixed(D) : '';     //  Total Errors (%). mean only for non-missing items in task
	sd = ( n > 1 ) ? Stat.sd(per_v, m).toFixed(D): '';  // Total Errors (%). standard deviation only for non-missing items in task  	 	
  rows.push( ['Total (%)', m, sd, n] );  // last row	
	data.push(m);   // index 2 => mean for percents
	data.push(sd);  // index 3 => stand. deviation for percents
	createTableArray(document.getElementById('table3'), rows, header); 

 
	// calculate inferential statistics (pearson r, sperman ro, t-arguments, t-distribution probabilities)	
	n= err_v.length;  // n= per_v.length; 
	header=  ['Variables', 'Spearman-ro', 't-value ro', 'df', 'p-ro (two-sided)',  'Pearson r', 't-value r', 'df', 'p-r (two-sided)' ]; // header 
	rows=[];    // reset table rows (for table4)
	for (let j= 0; j<=1; j++)
	{
		let y=[], vars;
		if ( j == 0 )  {
			y= err_v;   // error in pixels
			vars= var_name + '-Errors (px)';
		}
		else  {
			y= per_v;    // error in percents
			vars= var_name + '-Errors (%)';
		}	
		 
		let row= [ vars, '', '', '', '', '', '', '', '' ];  // range 0..8
		if  ( n >= 3 ) 
		{			
			let ro= Stat.r_spearman(var_v, y);     // spearman's rank correlation
			if (!isNaN(ro) && isFinite(ro) )  // if no variance in both arrays (sd= 0) => Infinity or Nan
			{ 
				row[1]= ro.toFixed(D);  
				let t_ro= ro * Math.sqrt ( ( n-2 ) / (1 - ro * ro) );  // t-argument for spearman correlation 		
        if (!isNaN(t_ro) && isFinite(t_ro))  
				{
					let err=[0];  // reset error flag indicator (must be 0)
					let p_ro= Stat.prncst ( Math.abs(t_ro), n-2, 0, err[0]); // t-argument, df, 0, error
					p_ro=  2 * (1 - p_ro);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[2]= t_ro.toFixed(D);
					row[3]= n-2;
					if (err[0] == 0) { 
						row[4]= p_ro.toFixed(D); // propability Spearman ro (two-sided)
					}
				}
			}
			
			let r= Stat.r(var_v, y);   // Pearson r correlation
			if (!isNaN(r) && isFinite(r))    // if no variance in both arrays (sd= 0) => Infinity or Nan
			{
				row[5]= r.toFixed(D);  
				let t_r=  r * Math.sqrt ( ( n-2 ) / (1 - r * r) );  // t-argument for pearson correlation
				if (!isNaN(t_r) && isFinite(t_r)) 
				{					
					let err=[0];  // reset error flag indicator (must be 0)
					let p_r= Stat.prncst ( Math.abs(t_r), n-2, 0, err[0]);  // t-argument, df, 0, error
					p_r= 2 * (1 - p_r);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[6]= t_r.toFixed(D);
					row[7]= n-2;
					if (err[0] == 0) { 
						row[8]= p_r.toFixed(D);  // p-r (two-sided)
					}					 
				}
			}			
		}  //  if ( n >= 3 ) 
		
		rows.push( row );  // 9 x columns
		
		data.push(row[5]);  // pearson r  (index 4 or 7)
		data.push(row[6]);  // t-value r  (index 5 or 8)
		data.push(row[8]);  // p-r (two-sided) => index (6 or 9)
	}		
	
	data.push(n - 2);   // df is equal for px and %. => or: err_v.length - 2 or: n= per_v.length - 2;
	
	
	createTableArray(document.getElementById('table4'), rows, header); 
  
	n= err_v.length;  // non-missing item count (error px)
	document.getElementById('inf_warning').style.display= n >= 3 ? 'block' : 'none';	
	
  // update 'score' label => above "Repeat" button
	switch (n)  {
		case 0: 
			document.getElementById('score').textContent= 'Descriptive statistics cannot be calculated. Missing items= ' + blank;
			break;
		case 1: 
			document.getElementById('score').textContent= 'Total average error(px)= ' + data[0] + '. Standard deviation cannot be calculated. Missing items= ' + blank;
			break;	
		default:
		document.getElementById('score').textContent= 'Total average error(px)= ' + data[0] + '. Standard Deviation= ' + data[1] + '. Missing items= ' + blank; 
	}
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	let var_values= '';
  for (let key in var_g)  {
    var_values += key + ';';
  }
	
	try 
  {
    if (store) {
			setStorage(var_name, var_values, blank, data);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
	
}



// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 var_name => variable name (string description)
	 var_values => list of variable values (always in px). separator in list is ";"  
	 blank => number of missing items in task 
	 data => descriptive and inferential statistics (index range: 0..10). for details see theEnd()  */ 
function setStorage(var_name, var_values, blank, data)  
{    
	let oldItems = JSON.parse(localStorage.getItem('pi-v1.0')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
      'variable' : var_name,
			'iteration' : iter,
			'angle' : variable == 1 ? var_values : prop[0].angle,
			'gap' : variable == 2 ? var_values : prop[0].gap,
			'rotation' : variable == 3 ? var_values : prop[0].rotation,
			'size' : variable == 4 ? var_values : prop[0].length,	
			'gap-type' : gap_type,	
			'pen-w': pen_width,
			'is-circle' : is_circle ? 1 : 0, 	
			'is-back' : is_back ? 1 : 0,
			'demo' : demo ? 1 : 0, 			
			'items' : item_cnt,
			'missing'  : blank, 
			
			'm-px'  : data[0], 	
			'sd-px' : data[1],
      'm-%'   : data[2], 	
			'sd-%'  : data[3],
			'r-px'  : data[4],
      't-px'  : data[5],
      'p-px'  : data[6],			
			'r-%'   : data[7],
      't-%'   : data[8],
      'p-%'   : data[9],			
			'df'    : data[10]
  };

  oldItems.push(newItem);

  localStorage.setItem('pi-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);  
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('pi-v1.0')  ?  JSON.parse(localStorage.getItem('pi-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('pi-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('pi-v1.0');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('pi-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
		case 2:
			tbl = document.getElementById('table3').getElementsByTagName('table')[0];  
			break;			
		default:
			tbl = document.getElementById('table4').getElementsByTagName('table')[0];  // note: id='table4' is html div element
			break;	
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}


// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes/static
// http://www.rosettacode.org/mw/index.php?title=Special%3ASearch&search=integration&go=Go&fulltext=1
class Stat 
{ 
		
	// return arithmetic mean (mean, average) of input array (arr)
  // no validation => input array must be in valid number format	
	static m(arr)  
	{	
		let len= arr.length, sum= 0;
		for (let i=0; i<len; i++) {
			sum+= arr[i];
		}
		return (sum/arr.length);
	}	


	// return Sample Standard Deviation (with Bessel's correction) => N-1
	// no validation => input array must be in valid number format
	// arr = array argument;  mean = average for input arr 
	static sd(arr, mean) 
	{
		let len= arr.length, sum= 0, res;
		for (let i=0; i<len; i++) {
			sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
		}
		res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
		return res;
	}
	
	
	// https://www.psychometrica.de/correlation.html
	// return Pearson correlation coefficient [ Pearson's r, Pearson product-moment correlation coefficient, bivariate correlation ]
	//  x, y = arrays 
	static r(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          //MeanAndStdDev(a, M1, SD1);
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    //MeanAndStdDev(b, M2, SD2);		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov;  // covariance
    }
		return ( cov / ( n * sd1 * sd2 ) );   // denominator = n-1 (Bessel's correction)		
	}
	
	
	// return covariance
  // x, y = arrays   
	static cov(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov  ;  // covarijanca
    }
		return cov;   // denominator = n-1 (Bessel's correction)		
	}


	// https://www.statisticshowto.com/probability-and-statistics/correlation-coefficient-formula/spearman-rank-correlation-definition-calculate/   
	// https://mathcracker.com/spearman-correlation-calculator
	// compute Spearman's rank correlation coefficient
	// x, y = arrays
	static r_spearman(x, y)
	{ 
		if (x.length != y.length) {
			return;  // undefined
		}
		const N = x.length;  // n is the number of observations
		let xr = x.slice();  // must be
		xr= this.rankify(xr, xr.length);   // get (tied) ranks for 1. array values
		
		let yr = y.slice();  // must be
		yr= this.rankify(yr, yr.length);	  // get (tied) ranks for 2. array values
		
		let mx= this.m(xr);  // mean rank for 1. array
		let my= this.m(yr);  // mean rank for 2. array
		
		let sxy= 0, sx= 0, sy= 0;
		for (let i= 0; i < N; i++)  
		{
			sxy+= ( xr[i] - mx ) * ( yr[i] - my); 
			sx+= Math.pow( ( xr[i] - mx ), 2);
			sy+= Math.pow( ( yr[i] - my ), 2);
		}
		sxy= sxy/N;
		sx= sx/N;
    sy= sy/N;
    let res = sxy / Math.sqrt(sx * sy);
		return res;	
	}
	
	
	// https://www.geeksforgeeks.org/rank-elements-array/
	// function to find tied ranks
	// A => input array;  n => input array length
	static rankify(A, n) 
	{	
		let R = [...Array(n)];  // Rank Vector
		// Sweep through all elements in A for each
		// element count the number of less than and
		// equal elements separately in r and s.
		for (let i = 0; i < n; i++) 
		{
			let r = 1, s = 1;
			for (let j = 0; j < n; j++) {
				if (j != i && A[j] < A[i]) r += 1;
				if (j != i && A[j] == A[i]) s += 1;
			}
			// Use formula to obtain rank
			R[i] = parseFloat(r + parseFloat(s - 1) / parseFloat(2));
		}
		//for (let i = 0; i < n; i++)  document.write(parseFloat(R[i]).toFixed(1) + "  ");
		return R;
	}
 
   
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.cpp
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html


	//****************************************************************************80

	static alnorm ( x, upper )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALNORM computes the cumulative density of the standard normal distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    17 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    David Hill,
	//    Algorithm AS 66:
	//    The Normal Integral,
	//    Applied Statistics,
	//    Volume 22, Number 3, 1973, pages 424-427.
	//
	//  Parameters:
	//
	//    Input, double X, is one endpoint of the semi-infinite interval
	//    over which the integration takes place.
	//
	//    Input, bool UPPER, determines whether the upper or lower
	//    interval is to be integrated:
	//    .TRUE.  => integrate from X to + Infinity;
	//    .FALSE. => integrate from - Infinity to X.
	//
	//    Output, double ALNORM, the integral of the standard normal
	//    distribution over the desired interval.
	//
	{
		let a1 = 5.75885480458;
		let a2 = 2.62433121679;
		let a3 = 5.92885724438;
		let b1 = -29.8213557807;
		let b2 = 48.6959930692;
		let c1 = -0.000000038052;
		let c2 = 0.000398064794;
		let c3 = -0.151679116635;
		let c4 = 4.8385912808;
		let c5 = 0.742380924027;
		let c6 = 3.99019417011;
		let con = 1.28;
		let d1 = 1.00000615302;
		let d2 = 1.98615381364;
		let d3 = 5.29330324926;
		let d4 = -15.1508972451;
		let d5 = 30.789933034;
		let ltone = 7.0;
		let p = 0.398942280444;
		let q = 0.39990348504;
		let r = 0.398942280385;
		let utzero = 18.66;
		let value;
		let y;
		let z = x;

		let up = upper;   // bool
		z = x;

		if ( z < 0.0 )
		{
			up = !up;
			z = - z;
		}

		if ( ltone < z && ( ( !up ) || utzero < z ) )
		{
			if ( up ) {
				value = 0;
			}
			else  {
				value = 1;
			}
			return value;
		}

		y = 0.5 * z * z;

		if ( z <= con )
		{
			value = 0.5 - z * ( p - q * y 
				/ ( y + a1 + b1 
				/ ( y + a2 + b2 
				/ ( y + a3 ))));
		}
		else
		{ 
			value = r * Math.exp( - y )   
				/ ( z + c1 + d1 
				/ ( z + c2 + d2 
				/ ( z + c3 + d3 
				/ ( z + c4 + d4 
				/ ( z + c5 + d5 
				/ ( z + c6 ))))));
		}

		if ( !up )  {
			value = 1 - value;
		}

		return value;
	}


	//****************************************************************************80

	static prncst ( st, idf, d, ifault )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    PRNCST computes the lower tail of noncentral T distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    26 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by BE Cooper.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    BE Cooper,
	//    Algorithm AS 5:
	//    The Integral of the Non-Central T-Distribution,
	//    Applied Statistics,
	//    Volume 17, Number 2, 1968, page 193.
	//
	//  Parameters:
	//
	//    Input, double ST, the argument.
	//
	//    Input, int IDF, the number of degrees of freedom.
	//
	//    Input, double D, the noncentrality parameter.
	//
	//    Output, int *IFAULT, error flag.
	//    0, no error occurred.
	//    nonzero, an error occurred.
	//
	//    Output, double PRNCST, the value of the lower tail of
	//    the noncentral T distribution.
	//
	//  Local Parameters:
	//
	//    Local, double G1, 1.0 / sqrt(2.0 * pi)
	//
	//    Local, double G2, 1.0 / (2.0 * pi)
	//
	//    Local, double G3, sqrt(2.0 * pi)
	//
	{
		let a, ak, b,  da, drb,  fk, fkm1, fmkm1, fmkm2, ioe, k, rb, sum, value;   
		let emin = 12.5;
		let g1 = 0.3989422804;
		let g2 = 0.1591549431;
		let g3 = 2.5066282746;
		let f = idf;   //( double ) ( idf );
	//
	//  For very large IDF, use the normal approximation.
	//
		if ( 100 < idf )
		{
			ifault = 1;

			a = Math.sqrt ( 0.5 * f ) * Math.exp ( this.lgamma ( 0.5 * ( f - 1 ) ) 
				- this.lgamma ( 0.5 * f ) ) * d;

			value = this.alnorm ( ( st - a ) / Math.sqrt ( f * ( 1 + d * d ) 
				/ ( f - 2 ) - a * a ), false );
			return value;
		}

		ifault = 0;
		ioe = ( idf % 2 );
		a = st / Math.sqrt ( f );
		b = f / ( f + st * st );
		rb = Math.sqrt ( b );
		da = d * a;
		drb = d * rb;

		if ( idf == 1 )
		{
			value = this.alnorm ( drb, true ) + 2 * this.tfn ( drb, a );
			return value;
		}

		sum = 0;

		if ( Math.abs ( drb ) < emin )
		{
			fmkm2 = a * rb * Math.exp ( - 0.5 * drb * drb ) 
			* this.alnorm ( a * drb, false ) * g1;
		}
		else
		{
			fmkm2 = 0;
		}

		fmkm1 = b * da * fmkm2;

		if ( Math.abs ( d ) < emin )
		{
			fmkm1 = fmkm1 + b * a * g2 * Math.exp ( - 0.5 * d * d );
		}

		if ( ioe == 0 )
		{
			sum = fmkm2;
		}
		else
		{
			sum = fmkm1;
		}

		ak = 1;
		fk = 2;

		for ( k = 2; k <= idf - 2; k = k + 2 )
		{
			fkm1 = fk - 1;
			fmkm2 = b * ( da * ak * fmkm1 + fmkm2 ) * fkm1 / fk;
			ak = 1 / ( ak * fkm1 );
			fmkm1 = b * ( da * ak * fmkm2 + fmkm1 ) * fk / ( fk + 1 );

			if ( ioe == 0 )
			{
				sum += fmkm2;
			}
			else
			{
				sum += fmkm1;
			}
			ak = 1 / ( ak * fk );
			fk += 2;
		}

		if ( ioe == 0 )
		{
			value = this.alnorm ( d, true ) + sum * g3;
		}
		else
		{
			value = this.alnorm ( drb, true ) + 2 * ( sum + this.tfn ( drb, a ) );
		}

		return value;
	}
	//****************************************************************************80


	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	//****************************************************************************80

	static lgamma ( x )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALOGAM computes the logarithm of the Gamma function.
	//
	//  Modified:
	//
	//    22 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by Malcolm Pike, David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    Malcolm Pike, David Hill,
	//    Algorithm 291:
	//    Logarithm of Gamma Function,
	//    Communications of the ACM,
	//    Volume 9, Number 9, September 1966, page 684.
	//
	//  Parameters:
	//
	//    Input, double X, the argument of the Gamma function.
	//    X should be greater than 0.
	//
	//    Output, double lgamma, the logarithm of the Gamma
	//    function of X.
	//
	{
		let f, value, y, z;

		if ( x <= 0 || isNaN(x) )  {
			return NaN;
		}

		y = x;

		if ( x < 7 )
		{
			f = 1;
			z = y;

			while ( z < 7 )
			{
				f = f * z;
				z++; 
			}
			y = z;
			f = - Math.log ( f );
		}
		else
		{
			f = 0;
		}

		z = 1 / y / y;

		value = f + ( y - 0.5 ) * Math.log ( y ) - y 
			+ 0.918938533204673 + 
			((( 
			- 0.000595238095238   * z 
			+ 0.000793650793651 ) * z 
			- 0.002777777777778 ) * z 
			+ 0.083333333333333 ) / y;

		return value;
	}

	//****************************************************************************80

	static tfn ( x, fx )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    TFN calculates the T-function of Owen.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    16 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by JC Young, Christoph Minder.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    MA Porter, DJ Winstanley,
	//    Remark AS R30:
	//    A Remark on Algorithm AS76:
	//    An Integral Useful in Calculating Noncentral T and Bivariate
	//    Normal Probabilities,
	//    Applied Statistics,
	//    Volume 28, Number 1, 1979, page 113.
	//
	//    JC Young, Christoph Minder,
	//    Algorithm AS 76: 
	//    An Algorithm Useful in Calculating Non-Central T and 
	//    Bivariate Normal Distributions,
	//    Applied Statistics,
	//    Volume 23, Number 3, 1974, pages 455-457.
	//
	//  Parameters:
	//
	//    Input, double X, FX, the parameters of the function.
	//
	//    Output, double TFN, the value of the T-function.
	//
	{
		const NG = 5;

		let fxs, i, r1, r2, rt, value, x1, x2, xs;
		let r = [
			0.1477621, 
			0.1346334, 
			0.1095432, 
			0.0747257, 
			0.0333357 ];
		let tp = 0.159155;
		let tv1 = 1.0E-35;
		let tv2 = 15.0;
		let tv3 = 15.0;
		let tv4 = 1.0E-05;
		let u = [
			0.0744372, 
			0.2166977, 
			0.3397048,
			0.4325317, 
			0.4869533 ];


	//
	//  Test for X near zero.
	//
		if ( Math.abs ( x ) < tv1 )
		{
			value = tp * Math.atan ( fx );
			return value;
		}
	//
	//  Test for large values of abs(X).
	//
		if ( tv2 < Math.abs ( x ) )
		{
			value = 0.0;
			return value;
		}
	//
	//  Test for FX near zero.
	//
		if ( Math.abs ( fx ) < tv1 )
		{
			value = 0;
			return value;
		}
	//
	//  Test whether abs ( FX ) is so large that it must be truncated.
	//
		xs = - 0.5 * x * x;
		x2 = fx;
		fxs = fx * fx;
	//
	//  Computation of truncation point by Newton iteration.
	//
		if ( tv3 <= Math.log ( 1 + fxs ) - xs * fxs )
		{
			x1 = 0.5 * fx;
			fxs = 0.25 * fxs;

			for ( ; ; )
			{
				rt = fxs + 1;

				x2 = x1 + ( xs * fxs + tv3 - Math.log ( rt ) ) 
				/ ( 2 * x1 * ( 1 / rt - xs ) );

				fxs = x2 * x2;

				if ( Math.abs ( x2 - x1 ) < tv4 )  {
					break;
				}
				x1 = x2;
			}
		}
	//
	//  Gaussian quadrature.
	//
		rt = 0.0;
		for ( i = 0; i < NG; i++ )
		{
			r1 = 1 + fxs * Math.pow ( 0.5 + u[i], 2 );
			r2 = 1 + fxs * Math.pow ( 0.5 - u[i], 2 );

			rt = rt + r[i] * ( Math.exp ( xs * r1 ) / r1 + Math.exp ( xs * r2 ) / r2 );
		}

		value = rt * x2 * tp;

		return value;

	}
	//****************************************************************************80
	
} 
